/*
    ************************************************
    * SIMULAZIONE COLLEGAMENTI AEREI RETE MONDIALE *
    *              con Algoritmo BFS               *
		* REALIZZATO DA:                               *
		* Marco Sommella 50/482                        *
		* Giovanni Di Cecca 50/887                     *
		* Virginia Bellino 408/466                     *
		************************************************
*/


#include "grafi.h"

/*
===========================
   FUNZIONI DI CREAZIONE
===========================
*/

/*	funzione per la creazione di un grafo con lista di adiacenza

   OUTPUT: ritorna il grafo creato
*/
graph *CreaGrafo ()
{
	graph *G = ( graph* ) malloc ( sizeof ( graph ) ); /* alloca il grafo */
	
    if ( G == NULL )
	{
		printf ( "\nERRORE: impossibile allocare memoria per il grafo\n" );
		system ( "PAUSE" );
	}
	else 	
    {
		G->ver = NULL;
		G->nv = 0;
		G->usati = 0;
	}

	return G;
}


/*	Funzione per la creazione di un vertice

		INPUT:  label - nome vertice
		        tipo - tipo di vertice
		OUTPUT: ritorna un nuovo vertice
*/
vertex *CreaVertice ( int label, int tipo )
{
	vertex *e = ( vertex * ) malloc ( sizeof ( vertex ) ); /* alloca un vertice */

  if ( e != NULL )
	{
		e->label = label;
		e->tipo = tipo;
		e->next = NULL;
	}

  return e;
}


/*
======================
   FUNZIONI DI TEST
======================
*/


/*	Controlla se il grafo  vuoto

		INPUT:	*G - puntatore al grafo
		OUTPUT:	ritorna 1 (vero) - il grafo  vuoto
		        ritorna 0 (falso) - il grafo non  vuoto
*/
int GrafoVuoto ( graph *G )
{
	return ( G == NULL );
}


/*	Trova una label nel grafo ritornando la posizione all'interno dell'array

		INPUT:	*G - puntatore al grafo
		        label - il nome cercato
		OUTPUT: ritorna -1 - label non trovata
						ritorna l'indice altrimenti
*/
int TrovaLabel ( graph *G, int label )
{
	int i=0, /* indice per il while */
			trovato=-1; /* variabile di ritorno */
	if ( !GrafoVuoto ( G ) )
	{
		while ( i < G->nv && trovato == -1 )
		{
			if ( G->ver[i] != NULL )
			{
				if ( G->ver[i]->label == label )
				{
					trovato = i;
				}
			}
			i++;
		}
	}
	return trovato;
}


/*	Controllo se esiste un arco tra u e v

INPUT: *G - puntatore al grafo
       u,v - indici dei vertici tra i quali si vuole verificare l'esistanza
  	        			di un arco
OUTPUT: ritorna 1 (vero) - adiacenti
        ritorna 0 (falso) - non adiacenti
*/
int Adiacenti ( graph *G, int u, int v )
{
	int trovato=0; /* variabile di ritorno */

  edge *e = G->ver[u]->next; /* puntatore per scorrere la lista di adiacenza */

  while ( e != NULL )
	{
		if ( e->key == v )
		{
			trovato = 1;
		}
		e = e->next;
	}

  return trovato;
}


/*	Verifica se u  vertice valido

  	INPUT:	*G - puntatore al grafo
  	        u - indici del vertice che si vuole verificare
    OUTPUT: ritorna 1 (vero) - valido
			ritorna 0 (falso) - non valido
*/
int VerticeValido ( graph *G, int u )
{
	if ( u != -1 )
	{
		if ( G->ver[u] != NULL )
		{
			return 1;
		}
	}
	else
	{
		return 0;
	}
}


/*	Trova l'HUB di un vertice locale u

		INPUT:	*G - puntatore al grafo
		        u - indice del vertice locale
		OUTPUT: ritorna -1 se non trovato
		        ritorna l'indice dell'HUB altrimenti
*/
int TrovaHUB ( graph *G, int u )
{
	int trovato=-1; /* variabile di ritorno */

	edge *e = G->ver[u]->next; /* puntatore per scorrere la lista di adiacenza */

	while ( e != NULL && trovato == -1 )
	{
		if ( e->tipo == 2 )
		{
			trovato = e->key;
		}
		e = e->next;
	}

	return trovato;
}


/*
=============================
   FUNZIONI DI INSERIMENTO
=============================
*/

/*  Inserisce un nuovo vertice nel grafo

		INPUT:	*G - puntatore al grafo
		        *vertice - puntatore al nuovo vertice
		OUTPUT: *G - puntatore al grafo modificato
*/
graph *InserisciVertice ( graph *G, vertex *vertice )
{
	vertex **e; /* puntatore a vertice */

	int i=0; /* indice per il while */

	if ( G->nv == 0 ) /* se non vi sono vertici si creano */
	{
		e = ( vertex ** ) realloc ( G->ver, ( G->nv + 1 ) * sizeof ( vertex * ) );
	}
	else
	{ /* si cerca una posizione dell'array non utilizzata */
		while (G->ver[i] != NULL && i < G->nv )  
		{
			i++;
		}
	}

	if ( i < G->nv ) /* se la posizione  trovata, la si usa*/
	{
		G->usati++;
		G->ver[i] = vertice;
	}
	else /* altrimenti si crea una nuova posizione */
	{
		e = ( vertex ** ) realloc ( G->ver, ( G->nv + 1 ) * sizeof ( vertex * ) );
		if ( e == NULL )
		{
			printf ( "\nERRORE: impossibile allocare memoria\n");
			system ( "PAUSE" );
		}
		else
		{
			G->ver = e;
			G->ver[G->nv] = vertice;
			G->nv++;
			G->usati++;
		}
	}

  return G;
}


/*  Inserisce un nuovo arco ( orientato ) nel grafo da u a v

		INPUT:	*G - puntatore al grafo
		        u,v - indici dei vertici tra i quali creare l'arco
*/
void InserisciArco ( graph *G, int u, int v )
{
	edge *e = ( edge * ) malloc ( sizeof ( edge ) ); /* alloca un nuovo arco */
	if ( e == NULL )
	{
  	printf ( "\nERRORE: impossibile allocare memoria\n");
		system ( "PAUSE" );
	}
	else
	{
		if ( !GrafoVuoto ( G ) )
		{
			if ( VerticeValido ( G, u ) && VerticeValido ( G, v ) )
			{
				if ( !Adiacenti ( G, u, v ) )
				{
					if ( G->ver[u]->tipo == 1 && G->ver[v]->tipo == 1 )
					{
						e->tipo = 1; /* si sta inserendo un LDA */
					}
					else
					{
    	  		e->tipo = 2; /* si sta inserendo un LNA */
					}
					e->key = v;
					e->next = G->ver[u]->next;
					G->ver[u]->next = e;
				}
			}
			else
			{
   			printf ( "\nERRORE: impossibile creare un arco se uno dei 2 vertici non esiste\n");
				system ( "PAUSE" );
			}
		}
		else
		{
   		printf ( "\nERRORE: impossibile creare un arco in un grafo vuoto\n");
			system ( "PAUSE" );
		}
	}
}


/*  Inserisce un nuovo arco ( non orientato )

		INPUT:	*G - puntatore al grafo
		        u,v - indici dei vertici tra i quali creare l'arco
*/
void InserisciArcoN_O ( graph *G, int u, int v )
{
  InserisciArco ( G, u, v);
  InserisciArco ( G, v, u);
}


/*
===============================
   FUNZIONI DI CANCELLAZIONE
===============================
*/

/*  Elimina un arco ( orientato ) dal grafo da u a v

		INPUT:	*G - puntatore al grafo
		        u,v - indici dei vertici tra i quali eliminare l'arco
*/
void EliminaArco ( graph *G, int u, int v )
{
	edge *prev, /* l'arco precedente a quello da togliere nella lista */
					*e; /* l'arco da togliere dalla lista */

	if ( !GrafoVuoto ( G ) )
	{
		if ( VerticeValido ( G, u ) && VerticeValido ( G, v ) )
		{
			if ( Adiacenti ( G, u, v ) )
			{
        e = G->ver[u]->next;
				if ( e->key == v ) /* se  l'elemento in testa */
				{
          G->ver[u]->next = e->next;
				}
				else
				{
					prev = e;
					while ( prev->next->key != v ) /* altrimenti lo cerco */
					{
						prev = prev->next;
					}
					e = prev->next;
					prev->next = e->next;
				}
			}
			else
			{
   			printf ( "\nERRORE: impossibile eliminare un arco se non esiste\n");
				system ( "PAUSE" );
			}
		}
		else
		{
   		printf ( "\nERRORE: impossibile eliminare un arco se uno dei 2 vertici non esiste\n");
			system ( "PAUSE" );
		}
	}
	else
	{
  	printf ( "\nERRORE: impossibile eliminare un arco in un grafo vuoto\n");
		system ( "PAUSE" );
	}

	free ( e );
}


/*  Elimina un arco ( non orientato ) dal grafo da u a v

		INPUT:	*G - puntatore al grafo
		        u,v - indici dei vertici tra i quali eliminare l'arco
*/
void EliminaArcoN_O ( graph *G, int u, int v )
{
  EliminaArco ( G, u, v );

  EliminaArco ( G, v, u );
}


/*  Elimina un un vertice dal grafo ( lasciando la posizione dell'array a NULL )

		INPUT:	*G - puntatore al grafo
		        u - indice del vertice da eliminare
*/
void EliminaVertice ( graph *G, int u )
{
  vertex *e; /* puntatore per memorizzare l'indirizzo del vertice da eliminare */

	edge *a; /* puntatore per scorrere la lista di adiacenza */

	if ( !GrafoVuoto ( G ) )
	{
		if ( VerticeValido ( G, u ) )
		{
      e = G->ver[u];
      a = G->ver[u]->next;
			if ( G->ver[u]->tipo == 2 ) /* u  un locale devo solo eliminare
			                               il rispettivo LNA */
			{
				EliminaArcoN_O ( G, u, TrovaHUB ( G, u ) );
				G->ver[u] = NULL;
			}
			else /* u  un HUB */
			{
				while ( a != NULL )
				{
					if ( a->tipo == 1 ) /* devo eliminare tutti gli LDA relativi */
					{
						EliminaArcoN_O ( G, u, a->key );
					}
					else /* e tutti i locali ad esso collegati */
					{
						EliminaVertice ( G, a->key );
					}
					a = G->ver[u]->next;
				}
			}
			free ( e );
			G->ver[u] = NULL;
			--G->usati;
		}
		else
		{
   		printf ( "\nERRORE: impossibile eliminare un vertice se non esiste\n");
			system ( "PAUSE" );
		}
	}
	else
	{
  	printf ( "\nERRORE: impossibile eliminare un vertice in un grafo vuoto\n");
		system ( "PAUSE" );
	}
}


/*
========================
   FUNZIONI DI STAMPA
========================
*/

/* Stampa informazioni sul grafo

		INPUT: *G - puntatore al grafo
*/
void StampaGrafo ( graph *G )
{
	edge *tmp; /* puntatore per scorrere la lista di adiacenza */

  int i; /* indice del for */

  system ( "CLS" );

  if( !GrafoVuoto ( G ) )
	{
    printf ( "Sono stampati solo gli Aereoporti HUB.\n" );
		printf ( "Tra parentesi quadre si indica il tipo di tratta: 1 se LDA, 2 se LNA.\n" );
		printf ( "\nCi sono %d Aereoporti ( HUB e Locali )\n\n", G->usati );
		for (i = 0; i < G->nv; i++ )
		{
			if( G->ver[i] != NULL && G->ver[i]->tipo == 1 )
			{
    		printf ( "%5d ->", G->ver[i]->label );
				tmp = G->ver[i]->next;
				while ( tmp != NULL )
				{
					printf ( " %5d-[%d]", G->ver[tmp->key]->label, tmp->tipo );
					tmp = tmp->next;
				}
				printf ( "\n" );
			}
		}
	}
	else
	{
		printf("Grafo Vuoto");
	}
	printf ( "\n" );

  system("PAUSE");
}


/*
========================
   FUNZIONI DI VISITA
========================
*/

/*	Visita in ampiezza

		INPUT: *G - puntatore al grafo
		       start - punto di partenza della visita
		OUTPUT: modifica il campo dist dei vertici del grafo con la distanza da start
*/
void Ampiezza ( graph *G, int start )
{
	edge *Succ;		/* puntatore di ricerca tra gli archi orientati */

  int	IndNodo,	/* posizione nell'array del nodo di partenza */
			IndEsame,	/* posizione nell'array del nodo in esame */
			Livello,	/* distanza dei nodi dal nodo base */
			*Fifo,		/* coda per gli indici dei figli da visitare */
			Front,		/* prossimo elemento da estrarre in Fifo[] */
			Rear,			/* casella libera per prossimo inserimento in coda */
			*Colori, *Pred, *Dist;	/* vettore dei colori e dei predecessori */

  Colori = ( int * ) malloc ( G->nv * sizeof ( int ) );

  Pred = ( int * ) malloc ( G->nv * sizeof ( int ) );

  Dist = ( int * ) malloc ( G->nv * sizeof ( int ) );

	/*  Azzera la flag 'Nodo Visitato' in ogni nodo */
  for (IndEsame = 0; IndEsame < G->nv; IndEsame ++)
	{
    Dist[IndEsame] = INT_MAX;

    Colori[IndEsame] = 0;

    Pred[IndEsame] = -1;
  }
  IndNodo = start;

	Fifo = ( int * ) malloc ( sizeof ( int ) * G->nv );

  Front = Rear = 0;

  Fifo [Rear ++] = IndNodo; /* il nodo base in coda */

  Colori[IndNodo] = 1;  /* marca 'visitato' */

  Dist[IndNodo] = 0;

	while (Front < Rear)
  {
		IndEsame = Fifo[Front ++]; /* estrae un nodo dalla coda */

		/* Stampa informazioni sul Nodo */

    /* inserisce in coda tutti i figli non visitati */
		Succ = G->ver[IndEsame]->next;
  	while ( Succ != NULL )
		{
			if ( Colori[Succ->key] == 0 )
			{
				Fifo [Rear ++] =  Succ->key;
				Colori[Succ->key] = 1;
				Pred[Succ->key] = IndEsame;
				Dist[Succ->key] = Dist[Pred[Succ->key]] + 1;
			}
    	Succ = Succ->next;
		}
  }
  free (Fifo);
  free (Colori);
	free (Pred);
  for (IndEsame = 0; IndEsame < G->nv; IndEsame ++)
	{
		if ( G->ver[IndEsame] != NULL )
		{
    	G->ver[IndEsame]->dist = Dist[IndEsame];
		}
	}
	free (Dist);
}
