/*
    ************************************************
    * SIMULAZIONE COLLEGAMENTI AEREI RETE MONDIALE *
    *           con Algoritmo di Dijkstra          *
		* REALIZZATO DA:                               *
		* Marco Sommella 50/482                        *
		* Giovanni Di Cecca 50/887                     *
		* Virginia Bellino 408/466                     *
		************************************************
*/

/* heap con la regole che il padre  minore dei figli */
#include "heap.h"


/*
=========================
   Funzioni dello HEAP
=========================
*/

/*	Figlio sinistro

		INPUT:  i - indice della radice
		OUTPUT: ritorna il figlio sinistro di i
*/
int left(int i)
{
	return 2*i+1;
}


/*	Figlio destro

		INPUT:  i - indice della radice
		OUTPUT: ritorna il figlio destro di i
*/
int right(int i)
{
	return 2*i+2;
}


/* Padre

		INPUT:  i - indice della radice
		OUTPUT: ritorna il padre di i
*/
int p(int i)
{
	return (i-1)/2;
}


/* Scambia la posizione nell'heap di 2 elementi

		INPUT:  *q - puntatore all'heap
		        i,j - indici degli elementi da scambiare
		OUTPUT: modifica l'heap
*/
void swap ( heap *q, int i, int j )
{
	heap_el tmp = q->data[i];
	q->data[i] = q->data[j];
	q->data[j] = tmp;
}


/* Ripristina la condizione di heap

		INPUT:  *q - puntatore all'heap
		        i - radice del sotto-albero dal quale far partire il controllo
		OUTPUT: modifica l'heap
*/
void Heapify ( heap *q, int i )
{
	int l,r,min; /* varibili per memorizzare i figli e il minimo */
	l = left(i);
	r = right(i);
	if ( l < q->size && q->data[l].dist < q->data[i].dist )
		min = l;
	else
		min = i;
	if ( r < q->size && q->data[r].dist < q->data[min].dist )
		min = r;
	if ( min != i )
	{
		swap ( q, i, min );
		Heapify ( q, min );
	}
}


/*  Costruisce l'heap

		INPUT:  *q - puntatore all'heap
		OUTPUT: modifica l'heap
*/
void BuildHeap ( heap *q )
{
	int i; /* indice per il for */
	for (i=q->size/2; i>=0; i--)
		Heapify ( q, i);
}


/* Controlla se l'heap  vuoto

		INPUT:  *q - puntatore all'heap
		OUTPUT: ritorna vero se vuoto
*/
int EmptyHeap ( heap *q )
{
	return !( q->size >= 1 );
}


/*  Estrae il minimo dall'heap

    INPUT:  *q - puntatore all'heap
    OUTPUT: ritorna il minimo estratto
*/
int GetMin ( heap *q )
{
	int tmp = q->data[0].pos;
	q->data[0].s = 1;
	swap ( q, 0, --q->size );
	Heapify( q, 0 );
	return tmp;
}


/*  Trova la posizione corrispondente ad un elemento del heap nel grafo

		INPUT:  *G - puntatore al grafo
		        *q - puntatore all'heap
		        start - valore cercato
		OUTPUT: ritorna la posizione
*/
int TrovaCorr (graph *G, heap *q, int start )
{
	int i; /* indice per il for */
	for ( i = 0; i < G->usati; i++ )
	{
		if ( q->data[i].pos == start )
		{
			return i;
		}
	}
}


/*
===========================
   Algoritmo di Dijkstra
===========================
*/
/*	Algoritmo di Dijkstra per il percorso minimo

		INPUT:  *G - puntatore al grafo
		        start - punto di partenza
		OUTPUT: puntatore ad array con le distanze
*/
int *Dijkstra ( graph *G, int start )
{
	int i, j = 0, /* indici */
			u; /* memorizza il minimo */
	edge *a; /* puntatore per scorrere la lista di adiacenza */

	                /* init */
	heap *q = ( heap * ) malloc ( sizeof ( heap ) );

  q->data = ( heap_el * ) calloc ( G->usati, sizeof ( heap_el ) );

	q->size = G->usati;

	for ( i = 0; i < G->nv; i++ ) /* scrive i valori da considerare */
	{
   	if ( G->ver[i] != NULL )
		{
			q->data[j].pos = i;
			if ( i == start )
			{
				q->data[j].dist = 0;
				q->data[j].s = 1;
			}
			else
			{
        q->data[j].dist = INT_MAX;
        q->data[j].s = 0;
			}
			j++;
		}
	}

	BuildHeap ( q );

	while ( EmptyHeap ( q ) );
  {
		u = GetMin ( q );
		a = G->ver[u]->next;
		i = TrovaCorr ( G, q, u );

    while ( a != NULL ) /* RELAX */
		{
			j = TrovaCorr ( G, q, a->key );
			if ( q->data[j].s == 0)
			{
				if ( q->data[i].dist + 1 < q->data[j].dist )
				{
					q->data[j].dist = q->data[i].dist + 1;
				}
			}
			a = a->next;
		}
	}

	/* trascrive le distanze nel array di ritorno */
	int *dist = ( int * ) calloc ( G->nv , sizeof ( int ) );

  for ( i = 0; i < G->usati; i++ )
	{
		dist[q->data[i].pos] = q->data[i].dist;
	}

  free (q->data);

  free ( q );

	return dist;
}


/*	Verifica la condizione di distanza tra 2 vertici usando Dijkstra

		INPUT:  *G - puntatore al grafo
		OUTPUT: modifica il grafo G se necessario
*/
void ControllaPercorsi ( graph *G )
{
	int i, j, k, z=-1;

	int *dist;

	for ( i = 0; i < G->nv; i++ )
	{
		if ( G->ver[i] != NULL )
		{
    	dist = Dijkstra ( G, i );
    	if ( z == -1 )
				z = i;
    	for ( j = 0; j < G->nv; j++ )
    	{
   			if ( dist[j] > 4 && G->ver[j] != NULL )
				{
					if ( G->ver[z]->tipo == 2 )
					{
						z = TrovaHUB ( G, z );
					}
					if ( G->ver[j]->tipo == 2 )
					{
						k = TrovaHUB ( G, j );
					}
					else
					{
						k = j;
					}
					if ( z != k )
					{
        		InserisciArcoN_O ( G, z, k );
					}
				}
			}
		}
	}
}
